@@ -39,12 +39,14 @@
 #include <bluetooth/sdp_lib.h>
 
 #include "att.h"
+#include "btio.h"
 #include "gattrib.h"
 #include "gatt.h"
 
 #define GATT_PSM 27
 /* Minimum MTU for L2CAP connections over BR/EDR */
 #define ATT_MIN_MTU_L2CAP 48
+#define GATT_CID 4
 
 static gchar *opt_src = NULL;
 static gchar *opt_dst = NULL;
@@ -56,8 +58,8 @@ static gboolean opt_primary = FALSE;
 static gboolean opt_characteristics = FALSE;
 static gboolean opt_char_read = FALSE;
 static gboolean opt_listen = FALSE;
-static guint listen_watch = 0;
 static gboolean opt_char_desc = FALSE;
+static gboolean opt_le = FALSE;
 static GMainLoop *event_loop;
 
 struct characteristic_data {
@@ -66,26 +68,30 @@ struct characteristic_data {
 	uint16_t end;
 };
 
-static int l2cap_connect(void)
+static void connect_cb(GIOChannel *io, GError *err, gpointer user_data)
 {
-	struct sockaddr_l2 addr;
-	bdaddr_t sba, dba;
-	int err, sk;
+	if (err)
+		g_printerr("%s", err->message);
+}
 
-	/* Remote device */
-	if (opt_dst == NULL) {
-		g_printerr("Remote Bluetooth address required\n");
-		return -EINVAL;
-	}
+static GIOChannel *do_connect(gboolean le)
+{
+	GIOChannel *chan;
+	bdaddr_t sba, dba;
 
 	/* This check is required because currently setsockopt() returns no
 	 * errors for MTU values smaller than the allowed minimum. */
 	if (opt_mtu != 0 && opt_mtu < ATT_MIN_MTU_L2CAP) {
 		g_printerr("MTU cannot be smaller than %d\n",
 							ATT_MIN_MTU_L2CAP);
-		return -EINVAL;
+		return NULL;
 	}
 
+	/* Remote device */
+	if (opt_dst == NULL) {
+		g_printerr("Remote Bluetooth address required\n");
+		return NULL;
+	}
 	str2ba(opt_dst, &dba);
 
 	/* Local adapter */
@@ -97,80 +103,22 @@ static int l2cap_connect(void)
 	} else
 		bacpy(&sba, BDADDR_ANY);
 
-	sk = socket(PF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_L2CAP);
-	if (sk < 0) {
-		err = errno;
-		g_printerr("L2CAP socket create failed: %s(%d)\n",
-							strerror(err), err);
-		return -err;
-	}
-
-	memset(&addr, 0, sizeof(addr));
-	addr.l2_family = AF_BLUETOOTH;
-	bacpy(&addr.l2_bdaddr, &sba);
-
-	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
-		err = errno;
-		g_printerr("L2CAP socket bind failed: %s(%d)\n",
-							strerror(err), err);
-		close(sk);
-		return -err;
-	}
-
-	if (opt_mtu > 0) {
-		struct l2cap_options l2o;
-		socklen_t len;
-
-		memset(&l2o, 0, sizeof(l2o));
-		len = sizeof(l2o);
-		if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &len) < 0) {
-			err = errno;
-			g_printerr("getsockopt(L2CAP_OPTIONS): %s(%d)\n",
-							strerror(err), err);
-			return -err;
-		}
-
-		l2o.imtu = opt_mtu;
-		l2o.omtu = opt_mtu;
-
-		if (setsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o,
-							sizeof(l2o)) < 0) {
-			err = errno;
-			g_printerr("setsockopt(L2CAP_OPTIONS): %s(%d)\n",
-							strerror(err), err);
-			return -err;
-		}
-	}
-
-	memset(&addr, 0, sizeof(addr));
-	addr.l2_family = AF_BLUETOOTH;
-	bacpy(&addr.l2_bdaddr, &dba);
-	addr.l2_psm = htobs(GATT_PSM);
-
-	err = connect(sk, (struct sockaddr *) &addr, sizeof(addr));
-	if (err < 0) {
-		err = errno;
-		g_printerr("L2CAP socket connect failed: %s(%d)\n",
-							strerror(err), err);
-		close(sk);
-		return -err;
-	}
-
-	return sk;
-}
-
-static GIOChannel *do_connect(void)
-{
-	GIOChannel *chan;
-	int sk;
-
-	sk = l2cap_connect();
-	if (sk < 0)
-		return NULL;
-
-	chan = g_io_channel_unix_new(sk);
-	g_io_channel_set_flags(chan, G_IO_FLAG_NONBLOCK, NULL);
-	g_io_channel_set_close_on_unref(chan, TRUE);
+	if (le)
+		chan = bt_io_connect(BT_IO_L2CAP, connect_cb, NULL, NULL, NULL,
+				BT_IO_OPT_SOURCE_BDADDR, &sba,
+				BT_IO_OPT_DEST_BDADDR, &dba,
+				BT_IO_OPT_CID, GATT_CID,
+				BT_IO_OPT_OMTU, opt_mtu,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
+				BT_IO_OPT_INVALID);
+	else
+		chan = bt_io_connect(BT_IO_L2CAP, connect_cb, NULL, NULL, NULL,
+				BT_IO_OPT_SOURCE_BDADDR, &sba,
+				BT_IO_OPT_DEST_BDADDR, &dba,
+				BT_IO_OPT_PSM, GATT_PSM,
+				BT_IO_OPT_OMTU, opt_mtu,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
+				BT_IO_OPT_INVALID);
 
 	return chan;
 }
@@ -239,30 +187,46 @@ done:
 
 static void events_handler(const uint8_t *pdu, uint16_t len, gpointer user_data)
 {
-	uint16_t handle, i;
+	GAttrib *attrib = user_data;
+	uint8_t opdu[ATT_MAX_MTU];
+	uint16_t handle, i, olen = 0;
 
 	handle = att_get_u16((uint16_t *) &pdu[1]);
 
 	switch (pdu[0]) {
 	case ATT_OP_HANDLE_NOTIFY:
-		g_print("attr handle = 0x%04x value: ", handle);
-		for (i = 3; i < len; i++)
-			g_print("%02x ", pdu[i]);
-
-		g_print("\n");
+		g_print("Notification handle = 0x%04x value: ", handle);
 		break;
 	case ATT_OP_HANDLE_IND:
+		g_print("Indication   handle = 0x%04x value: ", handle);
 		break;
+	default:
+		g_print("Invalid opcode\n");
+		return;
 	}
+
+	for (i = 3; i < len; i++)
+		g_print("%02x ", pdu[i]);
+
+	g_print("\n");
+
+	if (pdu[0] == ATT_OP_HANDLE_NOTIFY)
+		return;
+
+	olen = enc_confirmation(opdu, sizeof(opdu));
+
+	if (olen > 0)
+		g_attrib_send(attrib, opdu[0], opdu, olen, NULL, NULL, NULL);
 }
 
 static gboolean listen_start(gpointer user_data)
 {
 	GAttrib *attrib = user_data;
-	uint8_t events = ATT_OP_HANDLE_NOTIFY;
 
-	listen_watch = g_attrib_register(attrib, events, events_handler,
-								NULL, NULL);
+	g_attrib_register(attrib, ATT_OP_HANDLE_NOTIFY, events_handler,
+							attrib, NULL);
+	g_attrib_register(attrib, ATT_OP_HANDLE_IND, events_handler,
+							attrib, NULL);
 
 	return FALSE;
 }
@@ -458,7 +422,9 @@ static GOptionEntry gatt_options[] = {
 	{ "char-desc", 0, 0, G_OPTION_ARG_NONE, &opt_char_desc,
 		"Characteristics Descriptor Discovery", NULL },
 	{ "listen", 0, 0, G_OPTION_ARG_NONE, &opt_listen,
-		"Listen for notifications", NULL },
+		"Listen for notifications and indications", NULL },
+	{ "le", 0, 0, G_OPTION_ARG_NONE, &opt_le,
+		"Use Bluetooth Low Energy transport", NULL },
 	{ NULL },
 };
 
@@ -528,7 +494,7 @@ int main(int argc, char *argv[])
 		goto done;
 	}
 
-	chan = do_connect();
+	chan = do_connect(opt_le);
 	if (chan == NULL) {
 		ret = 1;
 		goto done;
@@ -545,8 +511,7 @@ int main(int argc, char *argv[])
 
 	g_main_loop_run(event_loop);
 
-	if (listen_watch)
-		g_attrib_unregister(attrib, listen_watch);
+	g_attrib_unregister_all(attrib);
 
 	g_main_loop_unref(event_loop);
 
